
// ******************************************************
// Functions to manipulate Analyze 7.5 images 
// (C) Jussi Tohka, 2005  (Based partly on the code provided by David Shattuck)
// Later modified by Ilya Eckstein.
// Laboratory of NeuroImaging, Department of Neurology, UCLA Medical School, CA, USA
// Institute of Signal Processing, Tampere University of Technology, Finland  
// Version 1.0, for internal distribution only.
// e-mail jussi.tohka@{tut.fi,loni.ucla.edu}
// *****************************************************

#ifndef ANALYZE_H
#define ANALYZE_H

#include <iostream>
#include <fstream>
#include <string.h>
//#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
using namespace std;



#define MAX_FILE_NAME   1024


#define DT_NONE                 0 
#define DT_UNKNOWN              0      /*Unknown data type*/ 
#define DT_BINARY               1      /*Binary (1 bit per voxel)*/ 
#define DT_UNSIGNED_CHAR				2      /*Unsigned character (8 bits per voxel)*/ 
#define DT_SIGNED_SHORT					4      /*Signed short (16 bits per voxel)*/ 
#define DT_SIGNED_INT           8      /*Signed integer (32 bits per voxel)*/ 
#define DT_FLOAT                16     /*Floating point (32 bits per voxel)*/ 
#define DT_COMPLEX              32     /*Complex (64 bits per voxel; 2 floating point numbers) */
#define DT_DOUBLE               64     /*Double precision (64 bits per voxel)*/ 
#define DT_RGB                  128    /* */
#define DT_ALL                  255    /* */ 

#ifdef   MAX
#undef   MAX
#endif
#define  MAX( x, y )  ( ((x) >= (y)) ? (x) : (y) )

#ifdef   MIN
#undef   MIN
#endif
#define  MIN( x, y )  ( ((x) <= (y)) ? (x) : (y) )


#define ROUND(x) (ceil((x) - 0.5))

#define ARE_EQUALF(x,y) (fabs((x) - (y)) < 0.0000001)   

#define FLOAT_SWAP(a,b) { register float t=(a);(a)=(b);(b)=t; }

struct AnalyzeHeader {
	AnalyzeHeader()
	{
		memset((char *)this,0,sizeof(*this));
		regular = 'r';
		sizeof_hdr=sizeof(*this);
	}
	int		sizeof_hdr;	/* For ANALYZE compatibility only */
	char		pad1[28];
	int		extents;	/* For ANALYZE compatibility only */
	char		pad2[2];
	char		regular;	/* For ANALYZE compatibility only */
	char		pad3;
	short int	dims;		/* For ANALYZE compatibility only */
	short int	x_dim;		/* AIR */
	short int	y_dim;		/* AIR */
	short int	z_dim;		/* AIR */
	short int	t_dim;		/* For ANALYZE compatibility only */
	char		pad4[20];
	short int	datatype;	/* For ANALYZE compatibility only */
	short int	bits;		/* AIR */
	char		pad5[6];
	float		x_size;		/* AIR */
	float		y_size;		/* AIR */
	float		z_size;		/* AIR */
	float		x_orig;		/* dws */
	float		y_orig;		/* dws */
	float		z_orig;		/* dws */
	char		pad6[36];	// was 48, now 12 for origin codes.
	int		glmax;		/* AIR */
	int		glmin;		/* AIR */
	char		descrip[80];	/* AIR (non-essential) */
	char		pad7[120];
};

struct AnalyzeImage {
  AnalyzeHeader header;
  float* data; // Internal data presented using floats 
  float minVal, maxVal; // Ilya added 
}; 

struct AnalyzeLabelImage {
  AnalyzeHeader header;
  unsigned char* data; // If memory is valuable
} ;




class AnalyzeOutStream
{
public: 
	AnalyzeOutStream(const AnalyzeHeader& header);
	~AnalyzeOutStream();

	int write(const AnalyzeImage* img, bool signedData, const char* filename,  bool overWrite = true);
	bool writeNextValue(float val, bool signedData = true);
	bool writeNextValue(const char* multiVal, int datasize);
  	bool substOutFile(const char* newFile, bool append);
	bool writeHeader(const char* filename);
	
protected: // data
	const AnalyzeHeader& m_header;
	ofstream m_ofile;
	int m_imgSize;
	int m_errorCode;
	int m_iCurPos;

//protected: // func
public: // TODO: hack!!!!
	void writeValue(float val, bool signedData);
};



void swapHeader(AnalyzeHeader &h);

inline void headername(char *dst, const char *src)
{
	strcpy(dst,src);
//	char *p = strrchr(dst,'.');
//	if (!p)
//	{
//		strcat(dst,".hdr");
//	}
//	else
//	{
//		strcpy(p,".hdr");
//	}

	char *p = dst + strlen(dst) - 4; // .img or .hdr
	if ('.' != *p)
	{
		strcat(dst,".hdr");
	}
	else
	{
		strcpy(p,".hdr");
	}
}

inline void imagename(char *dst, const char *src)
{
	strcpy(dst,src);
//	char *p = strrchr(dst,'.');
//	if (!p)
//	{
//		strcat(dst,".img");
//	}
//	else
//	{
//		strcpy(p,".img");
//	}

	char *p = dst + strlen(dst) - 4; // .img or .hdr
	if ('.' != *p)
	{
		strcat(dst,".img");
	}
	else
	{
		strcpy(p,".img");
	}
}

inline bool readHeader(const char *ifname, AnalyzeHeader& header)
{
	char hfname[MAX_FILE_NAME];
	headername(hfname,ifname);
	ifstream ifile(hfname);
	if (!ifile) return false;
	ifile.read((char *)&header,sizeof(AnalyzeHeader));
	return true;
}

inline bool writeHeader(const char *ofname, const AnalyzeHeader& header)
{
	char hfname[MAX_FILE_NAME];
	headername(hfname,ofname);
	ofstream ofile(hfname);
	if (!ofile) return false;
	ofile.write((const char *)&header,sizeof(AnalyzeHeader));
	return true;
}


inline void byteswap(unsigned short &X)
{
	unsigned short a = (unsigned char)(X>>8);
	unsigned char b = (unsigned char)X;
	X = (int(b)<<8)|a;
}

inline void byteswap(short &X)
{
	unsigned char a = (unsigned char)(X>>8);
	unsigned char b = (unsigned char)X;
	X = (int(b)<<8)|a;
}

inline void swapx(unsigned int &X)
{
	unsigned short a = X>>16;
	byteswap(a);
	unsigned short b = (unsigned short)(X);
	byteswap(b);
	X = (int(b)<<16)|a;
}

inline void swapf(unsigned int *X)
{
	unsigned short a = *X>>16;
	byteswap(a);
	unsigned short b = (unsigned short)(*X);
	byteswap(b);
	*X = (int(b)<<16)|a;
}

inline void swapx(int &X)
{
	unsigned short a = X>>16;
	byteswap(a);
	unsigned short b = (unsigned short)(X);
	byteswap(b);
	X = (int(b)<<16)|a;
}

inline void swapHeader(AnalyzeHeader &h)
{
	swapx(h.sizeof_hdr); // int		sizeof_hdr;
	//char		pad1[28];
	swapx(h.extents);		//int		extents;	/* For ANALYZE compatibility only */
	//char		pad2[2];
	//char		regular;	/* For ANALYZE compatibility only */
	//char		pad3;
	byteswap(h.dims); //short int	dims;		/* For ANALYZE compatibility only */
	byteswap(h.x_dim);//short int	x_dim;		/* AIR */
	byteswap(h.y_dim);//short int	y_dim;		/* AIR */
	byteswap(h.z_dim);//short int	z_dim;		/* AIR */
	byteswap(h.t_dim);//short int	t_dim;		/* For ANALYZE compatibility only */
	//char		pad4[20];
	byteswap(h.datatype);//short int	datatype;	/* For ANALYZE compatibility only */
	byteswap(h.bits);//short int	bits;		/* AIR */
	//char		pad5[6];
	swapf((unsigned int *)&h.x_size);//float		x_size;		/* AIR */
	swapf((unsigned int *)&h.y_size);//float		y_size;		/* AIR */
	swapf((unsigned int *)&h.z_size);//float		z_size;		/* AIR */
	swapf((unsigned int *)&h.x_orig);//float		x_size;		/* AIR */
	swapf((unsigned int *)&h.y_orig);//float		y_size;		/* AIR */
	swapf((unsigned int *)&h.z_orig);//float		z_size;		/* AIR */
	//char		pad6[48];
	swapx(h.glmax);//int		glmax;		/* AIR */
	swapx(h.glmin);//int		glmin;		/* AIR */
	//char		descrip[80];	/* AIR (non-essential) */
	//char		pad7[120];
};
 
inline bool byteSwapNecessary(AnalyzeHeader* h) {
  if( h->sizeof_hdr == 348) return false;
  else return true;
};  

int readImage(const char* filename,AnalyzeImage* img, bool signedData = true);
//int writeImage(const char* filename, AnalyzeImage* img, bool overwrite, bool signedData = true);
int readLabelImage(const char* filename,AnalyzeLabelImage* img);
int writeLabelImage(const char* filename,AnalyzeLabelImage* img,bool overwrite);

bool copyImage(AnalyzeImage* source,AnalyzeImage* target);
bool copyLabelImage(AnalyzeLabelImage* source,AnalyzeLabelImage* target);

bool newImage(AnalyzeImage* img, AnalyzeImage* alike);  
bool newLabelImage(AnalyzeLabelImage* img, AnalyzeImage* alike);

bool binaryMask(AnalyzeImage* img,AnalyzeLabelImage* mask,float tr); 
bool mask2Image(AnalyzeLabelImage* source, AnalyzeImage* target);


bool findClosestNonZero(AnalyzeImage*,int x,int y, int z,int* cx, int* cy, int* cz,float tr);
void averageFilterZeros(AnalyzeImage* img,AnalyzeImage* refImg,float tr,int times);

inline void thresholdImage(AnalyzeImage* img,float val) {
  int i,imgsize;
  imgsize = (img->header.x_dim)*(img->header.y_dim)*(img->header.z_dim); 
  for(i = 0;i< imgsize;i++) {
    if(img->data[i] > val) img->data[i] = 1;
    else img->data[i] = 0;
  }
}   

inline void freeImage(AnalyzeImage* img) 
{
  delete[] img->data;
};

inline void freeLabelImage(AnalyzeLabelImage *img) 
{
  delete[] img->data;
};

inline float getVoxelValue(AnalyzeImage* img,int x, int y, int z) 
{
  return img->data[x + y*(img->header.x_dim) + z*(img->header.y_dim)*(img->header.x_dim)];
};

inline float getSafeVoxelValue(AnalyzeImage* img,int x, int y, int z)
{
  if((x < 0) || (y < 0) || (z < 0) || (x > (img->header.x_dim - 1)) || 
     (y > (img->header.y_dim - 1)) || (z > (img->header.z_dim - 1))) 
    return(0);
  else  return img->data[x + y*(img->header.x_dim) + z*(img->header.y_dim)*(img->header.x_dim)];
}

inline void putVoxelValue(AnalyzeImage* img,int x, int y, int z, float val) 
{
  img->data[x + y*(img->header.x_dim) + z*(img->header.y_dim)*(img->header.x_dim)] = val;
};


inline unsigned char getLabelValue(AnalyzeLabelImage* img,int x, int y, int z) 
{
  return img->data[x + y*(img->header.x_dim) + z*(img->header.y_dim)*(img->header.x_dim)];
};

// this is the same as the previous one exept returns 0 if we are ouside the image

inline unsigned char getSafeLabelValue(AnalyzeLabelImage* img,int x, int y, int z)
{
  if((x < 0) || (y < 0) || (z < 0) || (x > (img->header.x_dim - 1)) || 
     (y > (img->header.y_dim - 1)) || (z > (img->header.z_dim - 1))) 
    return(0);
  else  return img->data[x + y*(img->header.x_dim) + z*(img->header.y_dim)*(img->header.x_dim)];
}

inline void putLabelValue(AnalyzeLabelImage* img,int x, int y, int z, unsigned char val) 
{
  img->data[x + y*(img->header.x_dim) + z*(img->header.y_dim)*(img->header.x_dim)] = val;
};

// inline double gaussian(double x,double mu,double sigma) {
// return(exp(-(pow((x - mu),2))/(2 * sigma * sigma))/(sqrt(2 * PI) * sigma))) };

inline float imageMax(AnalyzeImage* img,AnalyzeLabelImage* mask) 
{   
  int i,j,k;
  float maximum;
  
  maximum = FLT_MIN;
  for(i = 0;i < img->header.x_dim;i++) {
    for(j = 0;j < img->header.y_dim;j++) {
       for(k = 0;k < img->header.z_dim;k++) {
         if(getLabelValue(mask,i,j,k) > 0) {
           if(getVoxelValue(img,i,j,k) > maximum) maximum = getVoxelValue(img,i,j,k);
         }
       } 
    }
  }
  return(maximum);
};

inline unsigned char labelMax(AnalyzeLabelImage* img) 
{   
  int i,j,k;
  unsigned char maximum;
  
  maximum = 0;
  for(i = 0;i < img->header.x_dim;i++) {
    for(j = 0;j < img->header.y_dim;j++) {
       for(k = 0;k < img->header.z_dim;k++) {
         if(getLabelValue(img,i,j,k) > maximum) maximum = getLabelValue(img,i,j,k);
         
       } 
    }
  }
  return(maximum);
};

  
inline float imageMin(AnalyzeImage* img,AnalyzeLabelImage* mask) 
{   
  int i,j,k;
  float minimum;
  
  minimum = FLT_MAX;
  for(i = 0;i < img->header.x_dim;i++) {
    for(j = 0;j < img->header.y_dim;j++) {
       for(k = 0;k < img->header.z_dim;k++) {
         if(getLabelValue(mask,i,j,k) > 0) {
           if(getVoxelValue(img,i,j,k) < minimum) minimum = getVoxelValue(img,i,j,k);
         }
       } 
    }
  }
  return(minimum);
};

float imageKth(AnalyzeImage* img, AnalyzeLabelImage* mask, float percentage);


// 3D erosion with 3 x 3 x 3 structuring element

inline void erode3D(AnalyzeLabelImage* img) 
{
  int i,j,k,x,y,z;
  bool boundaryVoxel;
  AnalyzeLabelImage tmp;

  copyLabelImage(img,&tmp);
  for(i = 0;i < img->header.x_dim;i++) {
    for(j = 0;j < img->header.y_dim;j++) {
      for(k = 0;k < img->header.z_dim;k++) {
        if(getLabelValue(img,i,j,k) > 0.5) {
          boundaryVoxel = false;
          for(x = -1; x < 2; x++) {
            for(y = -1; y < 2; y++) {
              for(z = -1; z < 2; z++) {
                if(getSafeLabelValue(&tmp,i + x, j + y, k + z) < 0.5) {
		  boundaryVoxel = true;
                }
	      }
	    }
          }
	  if(boundaryVoxel) {
            putLabelValue(img,i,j,k,0);
          }
	}
      }
    }
  }
  freeLabelImage(&tmp);
}

void medianFilter3D(AnalyzeImage* img, AnalyzeLabelImage* mask, int windowX, int windowY, int windowZ);
void knnFilter(AnalyzeImage* img, AnalyzeLabelImage* mask, int windowLenX, int windowLenY, int windowLenZ, int knn);
  
inline char maxArg(float* pval,int n)
{
  float maximum;
  int i,index;
  
  maximum = pval[0];
  index = 0;
  for(i = 1;i < n;i++) {
    if(pval[i] > maximum) {
      index = i;
      maximum = pval[i];
    }
  }
  return( (char) index);
}


//////////////////////////////////////////////////////////////////////////
// AnalyzeOutStream inline functions
inline 
bool AnalyzeOutStream::writeNextValue(float val, bool signedData/* = true*/)
{
	if (m_iCurPos++ < m_imgSize)
	{
		writeValue(val, signedData);
		return true;
	}
	
	return false;
}


/*
 *	'datasize' must be in bytes 
 */
inline 
bool AnalyzeOutStream::writeNextValue(const char* multiVal, int datasize)
{
	if (m_header.datatype != DT_UNKNOWN ||
		m_header.bits != datasize<<3)// converting bytes to bits
		return false;

	if (m_iCurPos++ < m_imgSize)
	{
		m_ofile.write(multiVal, datasize);

		return true;
	}
	
	return false;
}


inline 
void AnalyzeOutStream::writeValue(float val, bool signedData)
{
	unsigned char charVal;
	unsigned short ushortVal;
	short shortVal;
	int intVal;
	int nWords = 0;
	
	switch(m_header.datatype) 
	{
	case DT_UNSIGNED_CHAR:
		val = ROUND(val);
		val = MIN(val,255);
		val = MAX(val,0);
		charVal = (unsigned char) val;
		m_ofile.write((char *)&charVal,sizeof(char));
		break;
		
	case DT_SIGNED_SHORT:
		if(signedData) 
		{
			val = ROUND(val);
			val = MIN(val,32768);
			val = MAX(val,-32767);
			shortVal = (short) val;
			m_ofile.write((char *)&shortVal,sizeof(short));
		}
		else {
			val = ROUND(val);
			val = MIN(val,65535);
			val = MAX(val,0);
			ushortVal = (unsigned short) val;
			m_ofile.write((char *)&ushortVal,sizeof(short));
		}
		break;
		
	case DT_SIGNED_INT:
		val = ROUND(val);
		val = MIN(val,2147000000);
		val = MAX(val,-2147000000);
		intVal = (int) val;
		m_ofile.write((char *)&intVal,sizeof(int));
		break;
		
	case DT_FLOAT:
		m_ofile.write((char *)&val,sizeof(float));
		break; 


	case DT_UNKNOWN:
		if (0 == m_header.bits % (8*sizeof(float))) // the number of bits/pix is a multiple of sizeof(float)
		{
			m_ofile.write((char *)&val,sizeof(float));
			break;
		}
	
		
	default: 
		;//ASSERT(false);
	}
}


#endif
